/****************************************************************************

Modified dates 
2007/11/22 Pascu Moca 
2008/05/08 Pascu Moca
2008/05/16 Pascu Moca
Documentation update necessary


*****************************************************************************/


#include "../include/real_axis_spectral_function.h"


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


double log_gauss_distribution(double omega, double omega_discrete, double b, double temperature){

/*******************************************************************************
input variables:  two
     omega          - the energy for which the spectral function is computed
     omega_discrete - the discrete value of the energy where delta peaks are computed
     b              - broadening parameter as in the case of Bulla's papers
     temperature    - is the temperature for which the calcuation is done. 

output variables: none

return: double that is broadening of the delta peak into a log-gaussian at a given energy. 

What it does:
     This routine computes the log gauss distribution for a given energy omega. 
    
How to call:
*******************************************************************************/

  double result; 

  if (b<= 0){
    error_message("b has to be positive\nError in routine log_gauss_distribution\nExitting..."); 
  }//if

  result=0; 

  /* if the temperature is 0 we use a lognormal function for the negative/positive frequency side */

  if (!(temperature>0))
    {
      if( ((omega>0)&& (omega_discrete>0)) || ((omega<0)&& (omega_discrete<0)) )
	{
	  result = exp (-0.25*b*b)/(b*fabs(omega_discrete)*sqrt(4.0*atan(1.0)))*exp(-pow((log(omega/omega_discrete)/b),2.0));
	} else 
	if ( omega * omega_discrete <= 0) {
	  result =0.0; 
	}

      return result;
    }//if ! temperature 
  

  /* if the temperature is non-zero, we  match  a lognormal and a Lorentzian  function */

  if (temperature>0){
    double log_weight =0.0 ;
    double lorentz_weight = 0.0 ;

    lorentz_weight = 2.0/atan(1.0) *0.1*b *temperature/(pow((omega-omega_discrete), 2)+pow (2.0*0.1*b*temperature,2)); 
    if (omega * omega_discrete >0) { 
      log_weight = exp (-0.25*b*b)/(b*fabs(omega_discrete)*sqrt(4.0*atan(1.0)))*exp(-pow((log(fabs(omega/omega_discrete))/b),2.0));
    }

    if (fabs(omega) < 4.0*temperature) {
      if (log_weight > 0) {
	lorentz_weight *= ( 1.0 - fabs(omega) /4.0/temperature) ;
	log_weight *=  (fabs(omega) /4.0/temperature) ;

	result = lorentz_weight + log_weight;
      } else {
	result = lorentz_weight;
      }      
    } else { 
      result = log_weight;
    } 
    return result; 
  }//if temperature>0;


}//log_gauss_distribution


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/



double log_gauss_distribution_modified (double omega, 
                                        double omega_discrete,                                          
                                        double quantum_temperature, 
                                        double b){

/*******************************************************************************
input variables:  five
     omega               - frequency at which interpolation is done 
     omega_discrete      - frequency of weight
     beta                - redundant 
     quantum_temperature - crossover scale from lognormal to Gaussian  
     b                   - parameter of lognormal function 

output variables: none

return: double that is ?????? 

What it does:
     ???
How to call: 
*******************************************************************************/

  double result;

  if(quantum_temperature==0){
    error_message("quantum_temperature is set to zero in the fitting_function");
  }
  double fitting_function_omega = (log(exp(1.1) + (pow(omega,2.0))/pow(quantum_temperature,2.0))*tanh(omega/quantum_temperature))/2.0;
  double fitting_function_omega_discrete = (log(exp(1.1) + (pow(omega_discrete/quantum_temperature,2.0)))
					    * tanh( omega_discrete/quantum_temperature))/2.0;

    double fitting_function_derivate_omega = (log(exp(1.1) + (pow(omega,2.0))/pow(quantum_temperature,2.0)))/
      (2.0*quantum_temperature*pow(cosh(omega/quantum_temperature),2.0))
      + omega*tanh(omega/quantum_temperature)/
      (pow (quantum_temperature, 2.0)*(exp(1.1)+(pow(omega,2.0))/pow(quantum_temperature,2.0)));


  result = 1.0/(b*sqrt(4*atan(1.0)))
    *exp(-pow((fitting_function_omega-fitting_function_omega_discrete), 2.0)/pow(b,2.0))
    *fitting_function_derivate_omega;

  return result;





  
}//log_gauss_distribution_modified



/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

long get_line_number(char* file_name) {
  fstream file_in;
  char error_string[400]; 


  
  strcpy(error_string,"cannot open the file: ");
  strcat(error_string,file_name);
  

  file_in.open(file_name, ios::in); 
  
  if (!file_in.good()){
    error_message(error_string);
  }//if
  
  /* Now we count the number of lines */  
  long line_no=0; 

  while(!(file_in.eof())){
    file_in.getline(buffer, buffer_size, '\n');
    line_no++; 
  }
  line_no--; 
  
  file_in.close();
  file_in.clear();

  return(line_no);
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void read_in_delta_peaks(char* file_name, 
			 long array_size,
			 double* E_array, 
			 double* delta_peaks_array, 
			 double &E_minimum, 
			 double &E_maximum)
{
  long no_of_lines; 
  fstream file_in;
  char error_string[400]; 

  strcpy(error_string,"cannot open the file: ");
  strcat(error_string,file_name);
  
  
  no_of_lines = get_line_number(file_name);
  
  if(no_of_lines > array_size) 
    error_message("no_of_lines > array_size in routine read_in_delta_peaks\n");


  /* open file containing delta peaks from iteration iteration_index */
  file_in.open(file_name, ios::in); 
  
  if (!file_in.good()){
    error_message(error_string);
  }//if


  for(int state_index=0; state_index<no_of_lines; state_index++){
  
     file_in.getline(buffer, buffer_size, '\n');
     word_offset = 0; 
      
      if(!(get_word(buffer, word, word_offset)))
        error_message("reading energies");
      E_array[state_index]= atof(word);     


      if(!(get_word(buffer, word, word_offset)))
        error_message("reading delta peaks");
      delta_peaks_array[state_index] = atof(word);

      /* remove matrix elements between states of the same energy */
      if(fabs(E_array[state_index]) < allowed_error) delta_peaks_array[state_index] = 0.0;      
      
  }// for state_index

   E_minimum  = E_array[0]; 
   E_maximum  = E_array[no_of_lines-1]; 

   file_in.close();

   return;
}




/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void reweight_spectral_function_peaks(quantum_operator & spectral_operator_A,
				      quantum_operator & spectral_operator_B,
				      char * positive_negative_spectrum_switch,
				      char * odd_even_iteration_switch, 
				      char * method
				      ){
  
/*******************************************************************************
input variables:  five
     spectral_operator_A - the first spectral operator in the combination <AB+> for which
                           the spectral function is computed.
     spectral_operator_B - the second spectral operator in the combination of the spectral function.

     positive_negative_spectrum_switch - switch for positive / negative part of the spectrum (the allowed values are 'positive' and 'negative' ).
     odd_even_iteration_switch - switch for odd/even iterations (allowed values are 
                                     'odd' or 'even' for method='NRG' 
                                      'odd_even' for method= 'DMNRG'
     method                - reweighting of the NRG or DMNRG spectrum (the allowed values are 'NRG' and 'DMNRG').
output variables: none

return: void

What it does:
     Reweights the spectral function delta peaks by combining the results from different iteration
     into a single file. Next these delta peaks will be bradened using the log-gauss distribution to 
     the real axis. 

How to call: 
*******************************************************************************/
  /*
  if ( strcmp(odd_even_iteration_switch,"odd") &&  
       strcmp(odd_even_iteration_switch,"even")&&
       strcmp(odd_even_iteration_switch, "odd_even")){
    error_message(" in the value of the odd_even_iteration_switch");
  }//if
  
  if ( strcmp(positive_negative_spectrum_switch,"positive") 
       &&  strcmp(positive_negative_spectrum_switch,"negative") ){
    error_message(" in the value of the positive_negative_spectrum_switch");
  }//if

   if ( strcmp(method, "NRG") && strcmp(method,"DMNRG") ){
    error_message(" in the value of the method");
  }//if
  */

 
  double * E_array; 
  double * coefficient_array; 
  double * matrix_element_array; 
  double * weight_array; 
  unsigned long * E_index_array; 
  int line_no; 
  long state_no; 
    
  double E_min[iteration_no+1];
  double E_max[iteration_no+1];
  double E[iteration_no+1][interval_no];
  double matrix_element[iteration_no+1][interval_no];
  double coefficient[iteration_no+1][interval_no];

  long no_of_nonzero_elements[iteration_no+1];
  double weight[iteration_no+1][interval_no];
 
  double sqrt_lambda = sqrt(lambda);
  double scale = 0.5 * (1 + 1 / lambda);
  int label; 



  /* checking if spectral functions' quantum numbers match */

  if(spectral_operator_A.get_representation_index()!= spectral_operator_B.get_representation_index()){
    error_message("spectral operators have not the same representation_index.\nEXITING...\n");
  }//if 
  if (spectral_operator_A.get_sgn()!= spectral_operator_B.get_sgn()){
    error_message("spectral operators have not the same signs.\nEXITING...\n");
  }//if

 

  /* create / open file for collected delta peaks from even iterations 
   construct the name of the output file: function  defined in in_out.cc / in_out.h 
   we open a new file when we generate the negative part of the spectrum */



  /* do the reweigting of the delta peaks  if 
     positive_negative_spectrum_switch = 'negative' and 
     odd_even_iteration_switch= 'even' and 
     method='NRG' 
  */
  
  if (!strcmp(positive_negative_spectrum_switch,"negative")&& 
      !strcmp(odd_even_iteration_switch,"even") &&
      !strcmp(method, "NRG")){
    delta_peaks_even_swap_file_name(delta_peaks_folder_name,
				    delta_peaks_even_file_name,
				    spectral_operator_A,
				    spectral_operator_B); 
    
    
    delta_peaks_even_file_out.open(delta_peaks_even_file_name, ios::out); 
    delta_peaks_even_file_out.precision(double_precision);
    
    
    if (!delta_peaks_even_file_out.good()){
      error_message("cannot open delta_peaks_even_file");
    }//if
    
  }//if 

  //we just reopen the file when we generate the positive part of the spectrum

  /* do the reweigting of the delta peaks  if 
     positive_negative_spectrum_switch = 'positive' and 
     odd_even_iteration_switch= 'even' and 
     method='NRG' 
  */

  if (!strcmp(positive_negative_spectrum_switch,"positive")&& 
      !strcmp(odd_even_iteration_switch,"even") &&
      !strcmp(method, "NRG")){
    delta_peaks_even_swap_file_name(delta_peaks_folder_name,
				    delta_peaks_even_file_name,
				    spectral_operator_A,
				    spectral_operator_B); 

    
    delta_peaks_even_file_out.open(delta_peaks_even_file_name, ios::app); 
    delta_peaks_even_file_out.precision(double_precision);

  
    if (!delta_peaks_even_file_out.good()){
      error_message("cannot open delta_peaks_even_file");
    }//if
    
  }//if 
 
  
  /* do the reweigting of the delta peaks  if 
     positive_negative_spectrum_switch = 'negative' and 
     odd_even_iteration_switch= 'odd' and 
     method='NRG' 
  */
  if (!strcmp(positive_negative_spectrum_switch,"negative")&& 
      !strcmp(odd_even_iteration_switch,"odd") &&
      !strcmp(method, "NRG")){
    delta_peaks_odd_swap_file_name(delta_peaks_folder_name,
				    delta_peaks_odd_file_name,
				    spectral_operator_A,
				    spectral_operator_B); 
    
    
    delta_peaks_odd_file_out.open(delta_peaks_odd_file_name, ios::out); 
    delta_peaks_odd_file_out.precision(double_precision);
    
    
    if (!delta_peaks_odd_file_out.good()){
      error_message("cannot open delta_peaks_odd_file");
    }//if
    
  }//if 

  //we just reopen the file when we generate the positive part of the spectrum


  /* do the reweigting of the delta peaks  if 
     positive_negative_spectrum_switch = 'positive' and 
     odd_even_iteration_switch= 'odd' and 
     method='NRG' 
  */
  if (!strcmp(positive_negative_spectrum_switch,"positive")&&
      !strcmp(odd_even_iteration_switch,"odd") &&
      !strcmp(method, "NRG")){
    delta_peaks_odd_swap_file_name(delta_peaks_folder_name,
				    delta_peaks_odd_file_name,
				    spectral_operator_A,
				    spectral_operator_B); 

    
    delta_peaks_odd_file_out.open(delta_peaks_odd_file_name, ios::app); 
    delta_peaks_odd_file_out.precision(double_precision);

  
    if (!delta_peaks_odd_file_out.good()){
      error_message("cannot open delta_peaks_odd_file");
    }//if
    
  }//if 
 

  /* do the reweigting of the delta peaks  if 
     positive_negative_spectrum_switch = 'negative' and 
     odd_even_iteration_switch, = 'odd_even'
     method='DMNRG' 
  */

  //if we use the dmnrg method we open the another file.

  if (!strcmp(positive_negative_spectrum_switch,"negative")&&       
      !(strcmp(odd_even_iteration_switch, "odd_even"))&&
      !strcmp(method, "DMNRG")){

    delta_peaks_dmnrg_swap_file_name(delta_peaks_folder_name,
				     delta_peaks_dmnrg_file_name,
				     spectral_operator_A,
				     spectral_operator_B); 
    
    
    delta_peaks_dmnrg_file_out.open(delta_peaks_dmnrg_file_name, ios::out); 
    delta_peaks_dmnrg_file_out.precision(double_precision);
    
     /* do the reweigting of the delta peaks  if 
     positive_negative_spectrum_switch = 'positive' and 
     odd_even_iteration_switch, = 'odd_even'
     method='DMNRG' 
  */
    if (!delta_peaks_dmnrg_file_out.good()){
      error_message("cannot open delta_peaks_dmnrg_file");
    }//if
    
  }//if 

  //we just reopen the file when we generate the positive part of the spectrum


  if (!strcmp(positive_negative_spectrum_switch,"positive")&&
      !(strcmp(odd_even_iteration_switch, "odd_even"))&&
      !strcmp(method, "DMNRG")){
    delta_peaks_dmnrg_swap_file_name(delta_peaks_folder_name,
				     delta_peaks_dmnrg_file_name,
				     spectral_operator_A,
				     spectral_operator_B); 

    
    delta_peaks_dmnrg_file_out.open(delta_peaks_dmnrg_file_name, ios::app); 
    delta_peaks_dmnrg_file_out.precision(double_precision);
    
  
    if (!delta_peaks_dmnrg_file_out.good()){
      error_message("cannot open delta_peaks_dmnrg_file");
    }//if
    
  }//if 
 
  
  /* clear variables */   
  for (int i=0; i<=iteration_no; i++){
    E_min[i] = E_max[i] = 0.0; /* the smallest / largest energy difference  at iteration i */
    for(int j=0; j<interval_no; j++){
      E[i][j] = 0.0; /* localtions of delta peaks from iteration i */
      matrix_element[i][j] = weight[i][j] = 0.0;  
    }//j
  }//i
  
  


  if (!strcmp(positive_negative_spectrum_switch,"negative") &&
      !strcmp(method, "NRG")){

    for (int iteration_index=0; iteration_index<=iteration_no; iteration_index++){
      
      /* construct names of the input files: function  defined in in_out.cc / in_out.h  */     
      delta_peaks_negative_swap_file_name(delta_peaks_folder_name,
					  iteration_index,
					  delta_peaks_negative_file_name,
					  spectral_operator_A,
					  spectral_operator_B); 
      
      no_of_nonzero_elements[iteration_index]=  get_line_number(delta_peaks_negative_file_name);
      
      
      read_in_delta_peaks(delta_peaks_negative_file_name, 
			  interval_no, 
			  E[iteration_index],
			  matrix_element[iteration_index], 
			  E_min[iteration_index], 
			  E_max[iteration_index]);

    }//for iteration_index;      
  } //if



  if (!strcmp(positive_negative_spectrum_switch,"positive")&&
      !strcmp(method, "NRG")){

      for (int iteration_index=0; iteration_index<=iteration_no; iteration_index++){
      
      /* construct names of the input files: function  defined in in_out.cc / in_out.h  */
      delta_peaks_positive_swap_file_name(delta_peaks_folder_name,
                                          iteration_index,
                                          delta_peaks_positive_file_name,
                                          spectral_operator_A,
                                          spectral_operator_B);

      no_of_nonzero_elements[iteration_index]=  get_line_number(delta_peaks_positive_file_name);


      read_in_delta_peaks(delta_peaks_positive_file_name,
                          interval_no,
                          E[iteration_index],
                          matrix_element[iteration_index],
                          E_min[iteration_index],
                          E_max[iteration_index]);

      }//for iteration_index;
    }//if




 if (!strcmp(positive_negative_spectrum_switch,"negative") &&
      !strcmp(method, "DMNRG")){

    for (int iteration_index=m_star; iteration_index<=iteration_no; iteration_index++){
      
      /* construct names of the input files: function  defined in in_out.cc / in_out.h  */     
      delta_peaks_negative_dmnrg_swap_file_name(delta_peaks_folder_name,
						iteration_index,
						delta_peaks_negative_dmnrg_file_name,
						spectral_operator_A,
						spectral_operator_B); 
      
      no_of_nonzero_elements[iteration_index]=  get_line_number(delta_peaks_negative_dmnrg_file_name);
     
      read_in_delta_peaks(delta_peaks_negative_dmnrg_file_name, 
			  interval_no, 
			  E[iteration_index],
			  matrix_element[iteration_index], 
			  E_min[iteration_index], 
			  E_max[iteration_index]);

    }//for iteration_index;      
  } //if




 if (!strcmp(positive_negative_spectrum_switch,"positive") &&
      !strcmp(method, "DMNRG")){

    for (int iteration_index=m_star; iteration_index<=iteration_no; iteration_index++){
      
      /* construct names of the input files: function  defined in in_out.cc / in_out.h  */     
      delta_peaks_positive_dmnrg_swap_file_name(delta_peaks_folder_name,
						iteration_index,
						delta_peaks_positive_dmnrg_file_name,
						spectral_operator_A,
						spectral_operator_B); 
      
      no_of_nonzero_elements[iteration_index]=  get_line_number(delta_peaks_positive_dmnrg_file_name);
      
      
      read_in_delta_peaks(delta_peaks_positive_dmnrg_file_name, 
			  interval_no, 
			  E[iteration_index],
			  matrix_element[iteration_index], 
			  E_min[iteration_index], 
			  E_max[iteration_index]);

    }//for iteration_index;      
  } //if


  /* now rescale the energies to get energies in units of D */
  

  for(int iteration_index= 0; iteration_index <= iteration_no; iteration_index++){

    double E_factor = pow(sqrt_lambda, 1-iteration_index);
    double scaling_factor = pow(sqrt_lambda,  1-iteration_index) * scale;


    E_max[iteration_index] *= scaling_factor ; 
    E_min[iteration_index] *= scaling_factor; 

    for(int state_index = 0; state_index < interval_no ;state_index++) { 
      E[iteration_index][state_index] *=  scaling_factor; 
      weight[iteration_index][state_index] = 1.0;

      if (iteration_index+2<=m_star){
	weight[iteration_index][state_index] = 0.0;
      }
    }//for
  } //for iteration_index




  int starting_iteration_index; 
  

  if (!strcmp(odd_even_iteration_switch,"even")&&
      !strcmp(method, "NRG")){
    if (div(m_star,2).rem ==0){
      /* mstar is even */ 
      starting_iteration_index = m_star; 
    }else{
      /* mstar is odd */
      starting_iteration_index = m_star-1;
    }//else
  }//if odd_even_iteration_switch


  if (!strcmp(odd_even_iteration_switch,"odd") &&
      !strcmp(method, "NRG")){
    if (div(m_star,2).rem ==1){
      /* mstar is odd */ 
      starting_iteration_index = m_star; 
    }else{
      /* mstar is even */
      starting_iteration_index = m_star-1;
    }//else
  }//if odd_even_iteration_switch

  if (!strcmp (method, "DMNRG")){
    starting_iteration_index = m_star;
  }


  // generate weighting factors to compensate overcounting 
  // we do the rescaling recursively when the NRG method is used. 

  if (!strcmp(method, "NRG")){
   
    for(int iteration_index = starting_iteration_index; iteration_index <= iteration_no-2; iteration_index += 2){
      
      /* first look for absolute minimum energy from all previous iterations */
           
      double abs_energy_min = 1.0e+10;     
      for(int inner_iteration_index = starting_iteration_index; inner_iteration_index <= iteration_index ; inner_iteration_index += 2){
	if (E_min[inner_iteration_index] < abs_energy_min) abs_energy_min = E_min[inner_iteration_index];      
      }//for inner_iteration_index;
      
      
      /* now we generate weights for the previous iterations. */
      
      for(int inner_iteration_index = starting_iteration_index; inner_iteration_index <= iteration_index ; inner_iteration_index += 2){
	for(int state_index =0; state_index<no_of_nonzero_elements[inner_iteration_index]; state_index++){
	  double Energy = E[inner_iteration_index][state_index];	
	  if(Energy <= E_max[iteration_index + 2]){	    
	    weight[inner_iteration_index][state_index] *=  (1. - (E_max[iteration_index + 2] - Energy) 
							    / (E_max[iteration_index + 2]-abs_energy_min));
	  }//if Energy	
	}//for state_index
      }//for inner_iteration_index;

      
      /* reweight the iteration_index+2 now */
      
      for(int state_index =0; state_index<no_of_nonzero_elements[iteration_index+2]; state_index++){      
	if (E[iteration_index+2][state_index]>abs_energy_min){
	  weight[iteration_index + 2][state_index] *= (1. - (E[iteration_index + 2][state_index]-abs_energy_min) 
						       /( E_max[iteration_index + 2]-abs_energy_min));	    	    
	}//if
      }//for state_index
      
      
      //if we are at the last iteration we rescale the weights... by a magic factor
      
      if(iteration_index+4 >iteration_no){
	for(int state_index =0; state_index<no_of_nonzero_elements[iteration_index]; state_index++){
	  weight[iteration_index][state_index] *= 1.0;
	  //weight[iteration_index][state_index] *=0.5;
	}//for state_index;
      }//if
      
    }//for iteration_index; 

    
  }//if method == NRG 




  //now we count how many non-zero matrix elements we have

  state_no =0; 

  if (!strcmp(method, "NRG")){    
    for(int iteration_index=starting_iteration_index; iteration_index<=iteration_no; iteration_index+=2)
      for(int state_index =0; state_index<no_of_nonzero_elements[iteration_index]; state_index++)
	if(matrix_element[iteration_index][state_index]*weight[iteration_index][state_index]!=0.0) 
	  state_no++;  
  }//if method == NRG
  
  if (!strcmp(method, "DMNRG")){    
    for(int iteration_index=starting_iteration_index; iteration_index<=iteration_no; iteration_index+=1)
      for(int state_index =0; state_index<no_of_nonzero_elements[iteration_index]; state_index++)
	if(matrix_element[iteration_index][state_index]*weight[iteration_index][state_index]!=0.0) 
	  state_no++;  
  }//if method == DMNRG
  
  
  E_array = new double [state_no]; 
  matrix_element_array = new double[state_no]; 
  weight_array = new double [state_no]; 
  E_index_array = new unsigned long [state_no]; 

  
  // fill in the arrays. 
  
  long  state_label =0; 

  if (!strcmp(method, "NRG")){ 
    for(int iteration_index=starting_iteration_index; iteration_index<=iteration_no; iteration_index+=2){
      for(int state_index =0; state_index<no_of_nonzero_elements[iteration_index]; state_index++){
	if(matrix_element[iteration_index][state_index]*weight[iteration_index][state_index]!=0.0){
	  
	  E_array[state_label] = E[iteration_index][state_index];
	  matrix_element_array[state_label]= matrix_element[iteration_index][state_index];
	  weight_array[state_label] = weight[iteration_index][state_index];
	  state_label++;
	}//if
      }//for state_index
    }//for iteration_index
  }//if method ==NRG
  
  if (!strcmp(method, "DMNRG")){ 
    for(int iteration_index=starting_iteration_index; iteration_index<=iteration_no; iteration_index+=1){
      for(int state_index =0; state_index<no_of_nonzero_elements[iteration_index]; state_index++){
	if(matrix_element[iteration_index][state_index]*weight[iteration_index][state_index]!=0.0){
	  
	  E_array[state_label] = E[iteration_index][state_index];
	  matrix_element_array[state_label]= matrix_element[iteration_index][state_index];
	  weight_array[state_label] = weight[iteration_index][state_index];
	  state_label++;
	}//if
      }//for state_index
    }//for iteration_index
  }//if method ==NRG
  

  //now we generate an energy index table
  generate_energy_index_table_mod(state_no,E_array, E_index_array); 
 




  //  //save the data into the file.

 
  if(!strcmp(positive_negative_spectrum_switch,"negative")&& 
     !strcmp(odd_even_iteration_switch,"even")&&
     !strcmp(method, "NRG")){
    for (int i=state_no-1; i>=0; i--){
      delta_peaks_even_file_out <<-E_array[(long)E_index_array[i]] <<"\t" 
				<< matrix_element_array[(long)E_index_array[i]]* weight_array[(long)E_index_array[i]]
				<<"\t"
				<< weight_array[(long)E_index_array[i]] <<endl;
    }//for i
  }//if positive_negative_switch
  
  
  
  if(!strcmp(positive_negative_spectrum_switch, "positive")&& 
     !strcmp(odd_even_iteration_switch,"even")&&
     !strcmp(method, "NRG")){
    for (int i=0; i<=state_no-1; i++){     
      delta_peaks_even_file_out <<E_array[(long)E_index_array[i]] <<"\t" 
				<< matrix_element_array[(long)E_index_array[i]]* weight_array[(long)E_index_array[i]]
				<<"\t"
				<< weight_array[(long)E_index_array[i]] <<endl;
    }//for i
  }//if positive_negative_switch
 
  
  
 
  if(!strcmp(positive_negative_spectrum_switch,"negative")&& 
     !strcmp(odd_even_iteration_switch,"odd")&& 
     !strcmp(method, "NRG")){
    for (int i=state_no-1; i>=0; i--){
      delta_peaks_odd_file_out <<-E_array[(long)E_index_array[i]] <<"\t" 
				<< matrix_element_array[(long)E_index_array[i]]* weight_array[(long)E_index_array[i]]
				<<"\t"
				<< weight_array[(long)E_index_array[i]] <<endl;
    }//for i
  }//if positive_negative_switch
  
  
  
  if(!strcmp(positive_negative_spectrum_switch,"positive")&& 
     !strcmp(odd_even_iteration_switch,"odd")&&
     !strcmp(method,"NRG")){
    for (int i=0; i<=state_no-1; i++){     
      delta_peaks_odd_file_out <<E_array[(long)E_index_array[i]] <<"\t" 
			       << matrix_element_array[(long)E_index_array[i]]* weight_array[(long)E_index_array[i]]
				<<"\t"
			       << weight_array[(long)E_index_array[i]] <<endl;
    }//for i
  }//if positive_negative_switch
 
 
  if(!strcmp(positive_negative_spectrum_switch,"negative")&& 
     !strcmp(method, "DMNRG")){
    for (int i=state_no-1; i>=0; i--){
      delta_peaks_dmnrg_file_out <<-E_array[(long)E_index_array[i]] <<"\t" 
				 << matrix_element_array[(long)E_index_array[i]]* weight_array[(long)E_index_array[i]]
				 <<"\t"
				 << weight_array[(long)E_index_array[i]] <<endl;
    }//for i
  }//if positive_negative_switch
   
  
  if(!strcmp(positive_negative_spectrum_switch,"positive")&& 
     !strcmp(method,"DMNRG")){
    for (int i=0; i<=state_no-1; i++){     
      delta_peaks_dmnrg_file_out <<E_array[(long)E_index_array[i]] <<"\t" 
			       << matrix_element_array[(long)E_index_array[i]]* weight_array[(long)E_index_array[i]]
				<<"\t"
			       << weight_array[(long)E_index_array[i]] <<endl;
    }//for i
  }//if positive_negative_switch
 



  
  
  delete [] E_array; 
  delete [] matrix_element_array; 
  delete [] weight_array; 
  delete [] E_index_array; 

  if (!strcmp(odd_even_iteration_switch,"even") &&
      !strcmp(method, "NRG")){
    delta_peaks_even_file_out.close();
  }//if
 


  if (!strcmp(odd_even_iteration_switch,"odd")&&
      !strcmp (method, "NRG")){
    delta_peaks_odd_file_out.close();
  }//if

 if (!strcmp (method, "DMNRG")){
   delta_peaks_dmnrg_file_out.close();
 }//if
 








 //for the dmnrg part of the calculation we save data comming from each iteration into separate files. 

 if (!strcmp(positive_negative_spectrum_switch, "negative")&&
     !strcmp(method, "DMNRG")){
   for(int iteration_index=starting_iteration_index; iteration_index<=iteration_no; iteration_index+=1){
     
    
     delta_peaks_dmnrg_swap_file_name(delta_peaks_folder_name,
				      iteration_index,
				      delta_peaks_dmnrg_file_name,
				      spectral_operator_A,
				      spectral_operator_B); 

     delta_peaks_dmnrg_file_out.open(delta_peaks_dmnrg_file_name, ios::out); 
     delta_peaks_dmnrg_file_out.precision(double_precision);

     if (!delta_peaks_dmnrg_file_out.good()){
       error_message("cannot open delta_peaks_dmnrg_file");
     }//if
     
     
     for(int state_index =no_of_nonzero_elements[iteration_index]-1; state_index>=0; state_index--){
       delta_peaks_dmnrg_file_out<<-E[iteration_index][state_index]<<"\t"
				 <<matrix_element[iteration_index][state_index]<<"\t"
				 <<weight[iteration_index][state_index]<<endl;

     }//for state_index
     delta_peaks_dmnrg_file_out.close();
   }//for iteration_index
 }//if method == DMNRG

 if (!strcmp(positive_negative_spectrum_switch, "positive")&&
     !strcmp(method, "DMNRG")){
   for(int iteration_index=starting_iteration_index; iteration_index<=iteration_no; iteration_index+=1){
     
    
     delta_peaks_dmnrg_swap_file_name(delta_peaks_folder_name,
				      iteration_index,
				      delta_peaks_dmnrg_file_name,
				      spectral_operator_A,
				      spectral_operator_B); 

     delta_peaks_dmnrg_file_out.open(delta_peaks_dmnrg_file_name, ios::app); 
     delta_peaks_dmnrg_file_out.precision(double_precision);

     if (!delta_peaks_dmnrg_file_out.good()){
       error_message("cannot open delta_peaks_dmnrg_file");
     }//if
     
     
     for(int state_index =0; state_index <no_of_nonzero_elements[iteration_index];  state_index++){
       delta_peaks_dmnrg_file_out<<E[iteration_index][state_index]<<"\t"
				 <<matrix_element[iteration_index][state_index]<<"\t"
				 <<weight[iteration_index][state_index]<<endl;
       
     }//for state_index
     delta_peaks_dmnrg_file_out.close();
   }//for iteration_index
 }//if method == DMNRG
 
 return; 

}//reweight_spectral_function_peaks


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/





/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void generate_broadening_mesh(){

  double energy_min; 
  double energy_max; 
  
  
  energy_max = broadening_energy_maxim;
  energy_min = broadening_energy_minim;
  
  
  assert(energy_min>0);
  assert(energy_max>0);
  
  
  broadening_energy_array = new double [2*broadening_grid_mesh];



  double energy_step = exp(log(energy_max/energy_min)/((double)(broadening_grid_mesh)));
  double energy = -energy_max*energy_step;
    
  for (int index=0; index<broadening_grid_mesh; index++){
    energy = energy/energy_step;
    broadening_energy_array[index] = energy; 
  }//for index;


  //here comes the positive part 

  energy = fabs(energy/energy_step);
  
 for (int index=broadening_grid_mesh; index<2*broadening_grid_mesh; index++){
    energy = energy*energy_step;
    broadening_energy_array[index] = energy; 
  }//for index;

  return; 
  }//generate_broadening_mesh;






/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void generate_spectral_function_real_axis(quantum_operator & spectral_operator_A, 
					  quantum_operator & spectral_operator_B, 
					  double b, 
					  double temperature, 
					  double quantum_temperature, 
					  char * odd_even_iteration_switch,
					  char * broadening_method,
					  char * method){
/*******************************************************************************
input variables:  eight
    spectral_operator_A - the first spectral operator for which the spectral function is computed.
    spectral_operator_B - the second spectral operator for which the spectral function is computed. 
    b                   -  broadening parameter used for the conversion of the delta peaks into log-gaussial distributions.
    temperature         - is the temperature for which the calculation is done. 
    quantum_temperature -
    odd_even_iteration_switch - switch for chosing the odd/even part of the spectrum. 
    broadening_method              - LOG_GAUSS or INTERPOLATIVE_LOG_GAUSS
    method             - NRG / DMNRG type of calculation. 
output variables: none

return: void

What it does:
    This routine generates the spectral function on the real axis for the combination <AB+> 
    by using only the even part of the spectrum. Odd part is used separately to generate a
    different spectral function. This routine uses the log-gauss distribution for the 
    broadening of the delta peaks. 

How to call:
*******************************************************************************/
 

  int line_no ; 
  int energy_no; 
  
  
  double * energies_array; 
  double * coefficients_array; 
  double spectral_function; 
  



 if ( strcmp(odd_even_iteration_switch,"odd") &&  
       strcmp(odd_even_iteration_switch,"even")&&
       strcmp(odd_even_iteration_switch, "odd_even")){
    error_message(" in the value of the odd_even_iteration_switch");
  }//if
  
  if ( strcmp(method,"NRG") &&  
       strcmp(method,"DMNRG") ){
    error_message(" in the value of the method");
  }//if
  
 if ( strcmp(broadening_method,"LOG_GAUSS") &&  
      strcmp(broadening_method,"INTERPOLATIVE_LOG_GAUSS") ){
    error_message(" in the value of the broadening_method");
  }//if
  


  if (!strcmp(odd_even_iteration_switch, "even")&& 
      !strcmp(method, "NRG")){  
    spectral_function_even_swap_file_name(working_folder_name,
					  b,
					  spectral_function_file_name,
					  spectral_operator_A, 
					  spectral_operator_B
					  );

    
    spectral_function_file_out.open(spectral_function_file_name, ios::out); 
    

    if (!spectral_function_file_out.good()){
      error_message("cannot open spectral_function_even_file_out"); 
    }//if
    spectral_function_file_out.precision(double_precision);
  
  }//if strcmp

  if (!strcmp(odd_even_iteration_switch, "odd")&&
      !strcmp(method, "NRG")){  
    spectral_function_odd_swap_file_name(working_folder_name,
					 b,
					 spectral_function_file_name,
					 spectral_operator_A, 
					 spectral_operator_B
					 );
    
    
    spectral_function_file_out.open(spectral_function_file_name, ios::out); 
    if (!spectral_function_file_out.good()){
      error_message("cannot open spectral_function_odd_file_out"); 
    }//if
    spectral_function_file_out.precision(double_precision);
  }//if strcmp

 if (!strcmp(method, "DMNRG")){  
    spectral_function_dmnrg_swap_file_name(working_folder_name,
					   b,
					   spectral_function_file_name,
					   spectral_operator_A, 
					   spectral_operator_B
					   );
    
    
    spectral_function_file_out.open(spectral_function_file_name, ios::out); 
    if (!spectral_function_file_out.good()){
      error_message("cannot open spectral_function_dmnrg_file_out"); 
    }//if
    spectral_function_file_out.precision(double_precision);
  }//if strcmp




  char delta_peaks_file_name[400];
  fstream delta_peaks_file; 


  if (!strcmp(odd_even_iteration_switch, "even")&&
      !strcmp(method, "NRG")){
  
    delta_peaks_even_swap_file_name(delta_peaks_folder_name,
				    delta_peaks_file_name,
				    spectral_operator_A,
				    spectral_operator_B); 
  }//if strcmp

  if (!strcmp(odd_even_iteration_switch, "odd")&&
      !strcmp(method, "NRG")){
  
    delta_peaks_odd_swap_file_name(delta_peaks_folder_name,
				   delta_peaks_file_name,
				   spectral_operator_A,
				   spectral_operator_B); 
  }//if strcmp

 

 if (!strcmp(method, "DMNRG")){
  
   delta_peaks_dmnrg_swap_file_name(delta_peaks_folder_name,
				    delta_peaks_file_name,
				    spectral_operator_A,
				    spectral_operator_B); 
 }//if strcmp


  //first we get the number of lines in the file.
  
  line_no = get_line_number(delta_peaks_file_name); 
  
  energy_no = line_no; 
  

  energies_array = new double [energy_no]; 
  coefficients_array =new double [energy_no]; 

  

  delta_peaks_file.open(delta_peaks_file_name, ios::in); 
  
  if (!delta_peaks_file.good()){
    error_message("cannot open delta_peaks_file");
  }//if
  
  
  for (int energy_index =0; energy_index<energy_no; energy_index++){
    
    delta_peaks_file.getline(buffer, buffer_size, '\n');
    word_offset = 0; 
      
    if(!(get_word(buffer, word, word_offset)))
      error_message("reading energies in the delta_peaks_even_file_in");
      energies_array[energy_index]= atof(word);     

      if(!(get_word(buffer, word, word_offset)))
        error_message("reading coefficients in the delta_peaks_even_file_in");
      coefficients_array[energy_index] = atof(word);

      if(!(get_word(buffer, word, word_offset)))
        error_message("reading weightings  in the delta_peaks_even_file_in");    


  }//for energy_index

  
  generate_broadening_mesh();

  
  double broadening_energy; 
  double energy; 
  double coefficient; 

  for (int spectral_function_index=0; spectral_function_index<2*broadening_grid_mesh; spectral_function_index++){
    spectral_function= 0.0;      
    broadening_energy = broadening_energy_array[spectral_function_index];

    for(int energy_index=0; energy_index<energy_no; energy_index++){
      energy = energies_array[energy_index];
      coefficient = coefficients_array[energy_index]; 

      if (!strcmp(broadening_method, "LOG_GAUSS")){
	spectral_function += coefficient*log_gauss_distribution(broadening_energy, energy, b, temperature); 
      }//if strcmp

      if (!  strcmp(broadening_method, "INTERPOLATIVE_LOG_GAUSS")){
	spectral_function += coefficient*log_gauss_distribution_modified(broadening_energy, energy, quantum_temperature, b); 
      }//if strcmp
      
    }//for energy_index
    spectral_function_file_out<<broadening_energy<<"\t"<<spectral_function<<endl;
  }//for spectral_function_index
  


  spectral_function_file_out.close();
  delta_peaks_file.close();

  delete [] energies_array;
  delete [] coefficients_array;
  delete [] broadening_energy_array; 

  return; 
}//generate_spectral_function_real_axis


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/






/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void generate_spectral_function_real_axis_dmnrg_each_iteration(quantum_operator & spectral_operator_A, 
							       quantum_operator & spectral_operator_B, 
							       double b, 
							       double temperature, 
							       double quantum_temperature, 		    
  							       char * broadening_method
							       ){
/*******************************************************************************
input variables:  eight
    spectral_operator_A - the first spectral operator for which the spectral function is computed.
    spectral_operator_B - the second spectral operator for which the spectral function is computed. 
    b                   -  broadening parameter used for the conversion of the delta peaks into log-gaussial distributions.
    temperature         - is the temperature for which the calculation is done. 
    quantum_temperature -
    odd_even_iteration_switch - switch for chosing the odd/even part of the spectrum. 
    broadening_method              - LOG_GAUSS or INTERPOLATIVE_LOG_GAUSS
    method             - NRG / DMNRG type of calculation. 
output variables: none

return: void

What it does:
    This routine generates the spectral function on the real axis for the combination <AB+> 
    by using only the even part of the spectrum. Odd part is used separately to generate a
    different spectral function. This routine uses the log-gauss distribution for the 
    broadening of the delta peaks. 

How to call:
*******************************************************************************/
 

  int line_no ; 
  int energy_no; 
  
  
  double * energies_array; 
  double * coefficients_array; 
  double spectral_function; 
  


  if ( !strcmp(broadening_method,"LOG_GAUSS") &&  
       !strcmp(broadening_method,"INTERPOLATIVE_LOG_GAUSS") ){
    error_message(" in the value of the broadening_method");
  }//if

  generate_broadening_mesh();
  
  for (int iteration_index = m_star; iteration_index<=iteration_no; iteration_index++){
    
    spectral_function_dmnrg_swap_file_name(spectral_functions_folder_name,
					   iteration_index,
					   b,
					   spectral_function_file_name,
					   spectral_operator_A, 
					   spectral_operator_B
					   );
    

    
    spectral_function_file_out.open(spectral_function_file_name, ios::out); 
    if (!spectral_function_file_out.good()){
      error_message("cannot open spectral_function_file_out"); 
    }//if
    
    
    
    
    
    
    delta_peaks_dmnrg_swap_file_name(delta_peaks_folder_name,
				     iteration_index,
				     delta_peaks_dmnrg_file_name,
				     spectral_operator_A,
				     spectral_operator_B); 
    

    
    line_no = get_line_number(delta_peaks_dmnrg_file_name); 
    
    energy_no = line_no; 
   
    energies_array = new double [energy_no]; 
    coefficients_array =new double [energy_no]; 

   
    
    delta_peaks_dmnrg_file_in.open(delta_peaks_dmnrg_file_name, ios::in); 
  
    if (!delta_peaks_dmnrg_file_in.good()){
      error_message("cannot open delta_peaks_dmnrg_file_in");
    }//if
    
    
    for (int energy_index =0; energy_index<energy_no; energy_index++){
      
      delta_peaks_dmnrg_file_in.getline(buffer, buffer_size, '\n');
      word_offset = 0; 
      
      if(!(get_word(buffer, word, word_offset)))
	error_message("reading energies in the delta_peaks_dmnrg_file_in");
      energies_array[energy_index]= atof(word);     
      
      if(!(get_word(buffer, word, word_offset)))
        error_message("reading coefficients in the delta_peaks_dmnrg_file_in");
      coefficients_array[energy_index] = atof(word);
      
      if(!(get_word(buffer, word, word_offset)))
        error_message("reading weightings  in the delta_peaks_dmnrg_file_in");    
    }//for energy_index

    

    double broadening_energy; 
    double energy; 
    double coefficient; 

    for (int spectral_function_index=0; spectral_function_index<2*broadening_grid_mesh; spectral_function_index++){
      spectral_function= 0.0;      
      broadening_energy = broadening_energy_array[spectral_function_index];
      
      for(int energy_index=0; energy_index<energy_no; energy_index++){
	energy = energies_array[energy_index];
	coefficient = coefficients_array[energy_index]; 
	
	if (!strcmp(broadening_method, "LOG_GAUSS")){
	  spectral_function += coefficient*log_gauss_distribution(broadening_energy, energy, b, temperature); 
	}//if strcmp
	
	if (!  strcmp(broadening_method, "INTERPOLATIVE_LOG_GAUSS")){
	  spectral_function += coefficient*log_gauss_distribution_modified(broadening_energy, energy, quantum_temperature, b); 
	}//if strcmp
	
      }//for energy_index
      spectral_function_file_out<<broadening_energy<<"\t"<<spectral_function<<endl;
    }//for spectral_function_index
  
  spectral_function_file_out.close();
  delta_peaks_dmnrg_file_in.close();

  delete [] energies_array;
  delete [] coefficients_array;
 
    
  }//for iteration_index
 
 delete [] broadening_energy_array; 

  return; 
}//generate_spectral_function_real_dmnrg_axis_each_iteration


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/







/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void generate_static_averages(quantum_operator & spectral_operator_A, 
			      quantum_operator & spectral_operator_B){

/*******************************************************************************
input variables:  two
    spectral_operator_A - the first spectral operator for which the static average is computed.
    spectral_operator_B - the second spectral operator for which the static_average is computed. 


output variables: none

return: void

What it does:
    
    This routine generates the static average of two operator <A+ B> by summing the data saved 
    in the swap folder where the contribution of each iteration has been computed.     


How to call:
*******************************************************************************/

  
  char static_average_file_name_in[400];
  char static_average_file_name_out[400]; 

  ifstream static_average_file_in;
  ofstream static_average_file_out; 

  double *iteration_average; 

  iteration_average = new double [iteration_no+1];


  static_average_spectral_operators_swap_file_name(static_operators_folder_name,
						   static_average_file_name_in,
						   spectral_operator_A,
						   spectral_operator_B);
  

  static_average_file_in.open(static_average_file_name_in, ios::in);
  

  if (!static_average_file_in.good()){
    error_message("unable to read in static_average_file_in.");
  }//if


  for (int iteration_index=iteration_no; iteration_index>=m_star; iteration_index--){
      
    static_average_file_in.getline(buffer, buffer_size, '\n');
    word_offset = 0; 
      
    if(!(get_word(buffer, word, word_offset)))
      error_message("error reading iteration_index in the generate_static_averages");
    int iteration = atoi(word);     
    
      
    if(!(get_word(buffer, word, word_offset)))
      error_message("error reading iteration_average in the generate_static_averages");
    iteration_average[iteration_index] = atof(word);
  }//for iteration_index




    static_average_spectral_operators_swap_file_name(working_folder_name,
						     static_average_file_name_out,
						     spectral_operator_A,
						     spectral_operator_B);
  

    static_average_file_out.open(static_average_file_name_out, ios::out);
  
    if (!static_average_file_out.good()){
      error_message("unable to write to static_average_file_out.");
    }//if
    
    
    double static_average; 
    double rescaled_temperature; 
   
    static_average_file_out.precision(double_precision);

    static_average_file_out<<"####################################################"<<endl;
    static_average_file_out<<"#\t<"<< spectral_operator_A.get_name()<<";"<< spectral_operator_B.get_name()<<">\n" ;
    static_average_file_out<<"####################################################"<<endl;
    
    static_average_file_out<<"# Iteration   Partial contribution";
    
    static_average_file_out<<endl; 
    
    for(int iteration_index=m_star; iteration_index<=iteration_no; iteration_index++){	
      static_average =0; 
      for(int i=m_star; i<=iteration_index; i++){
	static_average += iteration_average[i]; 
      }//for i	
      static_average_file_out<< iteration_index<<"\t"<<static_average<<endl;
    }//for iteration_index
    

    static_average_file_out<<"#####################################################"<<endl;
    static_average_file_out<<"#\tTemperature\t Average"<<endl;
    static_average_file_out<<"#####################################################"<<endl;

   
    static_average_file_out<<"\t"<< temperature<<"\t"<<static_average<<endl; 

   
      
    
    delete [] iteration_average; 
    
    static_average_file_out.close();
    static_average_file_out.clear();
    
    static_average_file_in.close();
    static_average_file_in.clear();
    
    return; 

}//generate_static_averages



void generate_static_averages(quantum_operator & static_operator_A ){

/*******************************************************************************
input variables:  two
    static_operator_A - the static operator for which the static average is computed.


output variables: none

return: void

What it does:
    
    This routine generates the static average of two operator <A+ B> by summing the data saved 
    in the swap folder where the contribution of each iteration has been computed.     


How to call:
*******************************************************************************/

  
  char static_average_file_name_in[400];
  char static_average_file_name_out[400]; 

  ifstream static_average_file_in;
  ofstream static_average_file_out; 

  double *iteration_average; 

  iteration_average = new double [iteration_no+1];

  
  static_average_static_operator_swap_file_name(static_operators_folder_name,
						 static_average_file_name_in,
						 static_operator_A
						 );

  

  static_average_file_in.open(static_average_file_name_in, ios::in);
  

  if (!static_average_file_in.good()){
    error_message("unable to read in static_average_file_in.");
  }//if


  for (int iteration_index=iteration_no; iteration_index>=m_star; iteration_index--){
      
    static_average_file_in.getline(buffer, buffer_size, '\n');
    word_offset = 0; 
      
    if(!(get_word(buffer, word, word_offset)))
      error_message("error reading iteration_index in the generate_static_averages");
    int iteration = atoi(word);     
    
      
    if(!(get_word(buffer, word, word_offset)))
      error_message("error reading iteration_average in the generate_static_averages");
    iteration_average[iteration_index] = atof(word);
  }//for iteration_index

  
  static_average_static_operator_swap_file_name(working_folder_name,
						 static_average_file_name_out,
						 static_operator_A
						 );
  

    static_average_file_out.open(static_average_file_name_out, ios::out);
  
    if (!static_average_file_out.good()){
      error_message("unable to write to static_average_file_out.");
    }//if
    
    

    double static_average; 
    double rescaled_temperature; 
   
    static_average_file_out.precision(double_precision);

    static_average_file_out<<"####################################################"<<endl;
    static_average_file_out<<"#\t<"<< static_operator_A.get_name()<<">\n" ;
    static_average_file_out<<"####################################################"<<endl;
    
    static_average_file_out<<"# Iteration   Partial contribution";
    
    static_average_file_out<<endl; 
    
    for(int iteration_index=m_star; iteration_index<=iteration_no; iteration_index++){	
      static_average =0; 
      for(int i=m_star; i<=iteration_index; i++){
	static_average += iteration_average[i]; 
      }//for i	
      static_average_file_out<< iteration_index<<"\t"<<static_average<<endl;
    }//for iteration_index
    

    static_average_file_out<<"#####################################################"<<endl;
    static_average_file_out<<"#\tTemperature\t Average"<<endl;
    static_average_file_out<<"#####################################################"<<endl;

   
    static_average_file_out<<"\t"<< temperature<<"\t"<<static_average<<endl; 

   
      
    
    delete [] iteration_average; 
    
    static_average_file_out.close();
    static_average_file_out.clear();
    
    static_average_file_in.close();
    static_average_file_in.clear();
    
    return; 

}//generate_static_averages

